package com.bomostory.sceneeditmodule

import android.content.Context
import android.graphics.Bitmap
import android.media.MediaMetadataRetriever
import android.util.Log
import com.bomostory.sceneeditmodule.screen.movie.MovieEditActivity.Companion.FPS
import com.bomostory.sceneeditmodule.basicdata.Music
import com.bomostory.sceneeditmodule.basicdata.Project
import com.bomostory.sceneeditmodule.screen.movie.MediaCodecMovieEncoder
import com.bomostory.sceneeditmodule.screen.movie.MovieEncoder
import com.bomostory.sceneeditmodule.utils.FileUtils
import com.example.exportmedia.MediaCallBack
import com.example.exportmedia.MediaHelper
import com.example.exportmedia.audio.AudioConcat
import com.example.exportmedia.audio.AudioLooper
import com.example.exportmedia.audio.AudioMixer
import com.example.exportmedia.data.AudioSource
import com.example.exportmedia.video.MovieMakerV2
import io.reactivex.*
import io.reactivex.Observable
import io.reactivex.schedulers.Schedulers
import pl.droidsonroids.gif.GifDrawable
import java.io.File
import java.util.*
import java.util.concurrent.Semaphore
import kotlin.collections.ArrayList

class SuperMovieMaker {

    fun make(
            context: Context,
            outputFile: File,
            project: Project,
            musics: List<Music>,
            scaleWidth: Int,
            scaleHeight: Int,
            mediaHelper: MediaHelper,
            movieEncoder: MovieEncoder? = MediaCodecMovieEncoder()): Observable<String> {
        return Observable.create<String> { emitter ->

            try {
                val beginTime = System.nanoTime()
                var timeAudioMixer = 0L
                var timeMovieMakerV2 = 0L
                var timeVideoDraw = 0L
                var timeVideoEncode = 0L

                val tmpMovieDirectory = File("${FileUtils.getProjectPath(project)}/${Config.MOVIE_TEMPORARY_FILE_FOLDER_NAME}/")
                if (!tmpMovieDirectory.exists() || !tmpMovieDirectory.isDirectory)
                    tmpMovieDirectory.mkdirs()
                val tmpVideoFile = File("${tmpMovieDirectory.absolutePath}/tmpVideo.mp4")
                val tmpAudioFile = File("${tmpMovieDirectory.absolutePath}/tmpAudio.mp4")

                if (tmpVideoFile.exists())
                    tmpVideoFile.delete()
                if (tmpAudioFile.exists())
                    tmpAudioFile.delete()
                if (outputFile.exists())
                    outputFile.delete()

                emitter.onNext("Ready")
                movieEncoder?.setVideoDimension(scaleWidth, scaleHeight)
                movieEncoder?.prepare(outputFile, FPS)

                var time: Long
                val frameDataList = generateFrameDataList(project)
                Log.d(this::class.java.simpleName, "frameDataList.size = ${frameDataList.size}")
                val semaphore = Semaphore(5)
                var generateCount = 0
                var completeCount = 0

                SceneDrawer.reset()
                movieEncoder?.start()
                val upstream = Flowable.create<Pair<Bitmap, FrameData>> ({
                    frameDataList.forEach { frameData ->
                        semaphore.acquire()
                        val scene = project.story!!.scenes[frameData.sceneIndex]
                        time = System.nanoTime()
                        val bitmap = SceneDrawer.drawScene(context, scene, frameData.x, scaleWidth, scaleHeight, frameData.millisecond)
                        timeVideoDraw += System.nanoTime() - time
                        generateCount++
                        it.onNext(Pair(bitmap!!, frameData))
                    }
                    it.onComplete()
                }, BackpressureStrategy.BUFFER)

                val disposable = upstream
                        .observeOn(Schedulers.newThread())
                        .subscribe ({
                            val bitmap = it.first
                            val frameData = it.second
                            time = System.nanoTime()
                            movieEncoder?.addFrame(bitmap!!, frameData.repeat)
                            timeVideoEncode += System.nanoTime() - time
                            bitmap.recycle()

                            completeCount++
                            emitter.onNext(String.format("%3.2f%%", completeCount * 100.0 / frameDataList.size))
                            semaphore.release()
                        }, {
                            it.printStackTrace()
                        }, {
                            emitter.onNext("Audio mixing")
                        })
                SceneDrawer.reset()
                movieEncoder?.finish()

                if (!emitter.isDisposed) {
                    val audioSources = ArrayList<AudioSource>()
                    project.story?.let {
                        audioSources.add(generateRecordSource(project, mediaHelper))
                    }
                    audioSources.addAll(generateAudioSource(project, musics, mediaHelper))

                    time = System.nanoTime()
                    val audioMixerBuilder = AudioMixer.Builder(mediaHelper)
                    audioMixerBuilder.audioSourceList = audioSources
                    val audioMixer = audioMixerBuilder.build()
                    Observable.create<String> {
                        audioMixer.mediaCallBack = object: MediaCallBack {
                            override fun onExecuteSuccess(message: String?) {
                                message?.let { msg ->
                                    it.onNext(msg)
                                }
                            }

                            override fun onExecuteFailed(message: String?) {
                                it.onError(Throwable(message))
                            }

                            override fun onExecuteFinish() {
                                it.onComplete()
                            }
                        }
                        audioMixer.output(tmpAudioFile)
                    }.blockingSubscribe()
                    timeAudioMixer = System.nanoTime() - time
                    emitter.onNext("Encoding video")

                    val outputPath = outputFile.absolutePath
                    outputFile.renameTo(tmpVideoFile)

                    time = System.nanoTime()
                    val movieMakerBuilder = MovieMakerV2.Builder(mediaHelper)
                    movieMakerBuilder.videoPath = tmpVideoFile.absolutePath
                    movieMakerBuilder.audioPath = tmpAudioFile.absolutePath
                    val movieMaker = movieMakerBuilder.build()
                    Observable.create<String> {
                        movieMaker.mediaCallBack = object: MediaCallBack {
                            override fun onExecuteSuccess(message: String?) {
                                message?.let { msg ->
                                    it.onNext(msg)
                                }
                            }

                            override fun onExecuteFailed(message: String?) {
                                it.onError(Throwable(message))
                            }

                            override fun onExecuteFinish() {
                                it.onComplete()
                            }
                        }
                        movieMaker.output(File(outputPath))
                    }.blockingSubscribe()
                    timeMovieMakerV2 = System.nanoTime() - time
                }

                val endTime = System.nanoTime()
                Log.d(this::class.java.simpleName, "duration = ${(endTime - beginTime) / 1000000000.0}")
                Log.d(this::class.java.simpleName, "timeVideoDraw = ${timeVideoDraw / 1000000000.0}")
                Log.d(this::class.java.simpleName, "timeVideoEncode = ${timeVideoEncode / 1000000000.0}")
                Log.d(this::class.java.simpleName, "timeAudioMixer = ${timeAudioMixer / 1000000000.0}")
                Log.d(this::class.java.simpleName, "timeMovieMakerV2 = ${timeMovieMakerV2 / 1000000000.0}")

                tmpMovieDirectory.deleteRecursively()

                emitter.onComplete()
            } catch (e: Exception) {
                e.printStackTrace()
            }
        }
                .doOnDispose {
                    SceneDrawer.reset()
                    movieEncoder?.finish()
                }
    }

    private fun generateRecordSource(project: Project, mediaHelper: MediaHelper): AudioSource {
        val audioSource = AudioSource()

        val tmpMovieDirectory = File("${FileUtils.getProjectPath(project)}/${Config.MOVIE_TEMPORARY_FILE_FOLDER_NAME}/")
        val fileName = "tmpRecord${System.currentTimeMillis()}.mp3"
        val outputFile = File(tmpMovieDirectory, fileName)

        val audioConcatBuilder = AudioConcat.Builder(mediaHelper)
        project.story?.let {
            for (scene in it.scenes) {
                scene.recordPath?.let {
                    audioConcatBuilder.addInputSource(File(it))
                }
            }
        }

        val audioConcat = audioConcatBuilder.build()
        audioConcat.output(outputFile).blockingFirst()

        val mediaMetadataRetriever = MediaMetadataRetriever()
        mediaMetadataRetriever.setDataSource(outputFile.path)

        audioSource.path = outputFile.path
        audioSource.duration =  mediaMetadataRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_DURATION).toLong()
        return audioSource
    }

    private fun generateAudioSource(project: Project, musics: List<Music>, mediaHelper: MediaHelper): ArrayList<AudioSource> {
        val audioSources = ArrayList<AudioSource>()

        val tmpMovieDirectory = File("${FileUtils.getProjectPath(project)}/${Config.MOVIE_TEMPORARY_FILE_FOLDER_NAME}/")
        for (music in musics) {
            val audioSource = AudioSource()

            if (music.isLoop) {
                val fileName = "tmpMusic${System.currentTimeMillis()}.mp3"
                val outputFile = File(tmpMovieDirectory, fileName)

                val audioLooperBuilder = AudioLooper.Builder(mediaHelper)
                audioLooperBuilder.loopAudioFile = File(music.path)
                audioLooperBuilder.loopDuration = music.loopDuration.toInt()

                val audioLooper = audioLooperBuilder.build()
                audioLooper.output(outputFile).blockingFirst()

                val mediaMetadataRetriever = MediaMetadataRetriever()
                mediaMetadataRetriever.setDataSource(outputFile.path)

                audioSource.path = outputFile.path
                audioSource.duration = mediaMetadataRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_DURATION).toLong()
                audioSource.offsetSec = music.offsetTime.toFloat() / 1000
            } else {
                audioSource.path = music.path
                audioSource.duration = music.duration
                audioSource.offsetSec = music.offsetTime.toFloat() / 1000
            }
            audioSources.add(audioSource)
        }
        return audioSources
    }

    private fun generateFrameDataList(project: Project): List<FrameData> {
        val frameDataList = ArrayList<FrameData>()
        var index = 0
        project.story?.scenes?.forEach { scene ->
            val sceneIndex = index++

            scene.record?.apply {

                frameDataList.add(FrameData(sceneIndex, 0, 1L, 0L))

                var x = 0
                var trackPosition = -1

                val timeSet = TreeSet<Long>()
                for (layer in scene.layers) {
                    for (actor in layer.actors) {
                        if (actor.resourcePath.toLowerCase().endsWith(".gif")) {
                            val gifDrawable = GifDrawable(actor.resourcePath)
                            val numberOfFrames = gifDrawable.numberOfFrames
                            val loopCount = if (gifDrawable.loopCount == 0) Int.MAX_VALUE else gifDrawable.loopCount
                            val durationCount = Math.ceil(period / gifDrawable.duration.toDouble()).toInt()
                            var duration = 0L
                            for (times in 1..Math.min(loopCount, durationCount)) {
                                for (frameIndex in 0 until numberOfFrames) {
                                    duration += gifDrawable.getFrameDuration(frameIndex)
                                    if (duration > period)
                                        break
                                    timeSet.add(duration)
                                }
                                if (duration > period)
                                    break
                            }
                        }
                    }
                }

                val timeArray = timeSet.toArray() as Array<Any>
                var timeArrayIndex = 0

                for (t in 0..period step ((1f / FPS) * 1000).toLong()) {

                    //  gif animation
                    var gifUpdate = false
                    if (timeArrayIndex < timeArray.size) {
                        var gifTime = timeArray[timeArrayIndex] as Long
                        if (t >= gifTime) {
                            while (timeArrayIndex < timeArray.size && t > timeArray[timeArrayIndex] as Long) {
                                timeArrayIndex++
                            }
                            gifUpdate = true
                        }
                    }

                    var tmpTrackPosition = trackPosition
                    for (i in (trackPosition + 1) until tracks.size) {
                        val track = tracks[i]
                        if (t >= track.time) {
                            x = track.positionX
                            tmpTrackPosition = i
                        } else {
                            break
                        }
                    }
                    trackPosition = tmpTrackPosition

                    val frameData = frameDataList.last()
                    if (!gifUpdate && frameData.x == x) {
                        frameData.repeat++
                    } else {
                        frameDataList.add(FrameData(sceneIndex, x, 1L, t))
                    }
                }
            }
        }

        return frameDataList
    }

    data class FrameData(val sceneIndex: Int, val x: Int, var repeat: Long, val millisecond: Long)
}