package com.bomostory.sceneeditmodule.screen.movie

import android.arch.lifecycle.MutableLiveData
import android.arch.lifecycle.ViewModel
import android.content.ContentResolver
import android.content.Context
import android.media.MediaMetadataRetriever
import com.bomostory.sceneeditmodule.Config
import com.bomostory.sceneeditmodule.basicdata.*
import com.bomostory.sceneeditmodule.utils.FileUtils
import com.bomostory.sceneeditmodule.utils.MusicPlayerV2
import com.bomostory.sceneeditmodule.utils.TimeUtils
import com.example.bomocloud.BoMoCloud
import com.example.bomocloud.video.UploadVideoData
import com.example.exportmedia.MediaHelper
import com.example.exportmedia.audio.AudioEditor
import com.example.tfat.myapplication.R
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.schedulers.Schedulers
import okhttp3.MediaType
import okhttp3.MultipartBody
import okhttp3.RequestBody
import java.io.File
import java.util.concurrent.CopyOnWriteArrayList

class MovieEditViewModel : ViewModel() {

    companion object {
        val onProjectSavedListeners = CopyOnWriteArrayList<Runnable>()
    }

    val storyLiveData = MutableLiveData<Story>()

    val storyPeriodLiveData = MutableLiveData<Long>()

    val sceneLiveData = MutableLiveData<Scene>()

    val musicLiveData = MutableLiveData<Music>()

    val musicPlayPosition = MutableLiveData<Int>()

    val showProgressingLiveData = MutableLiveData<Boolean>()

    val musicListLiveData = MutableLiveData<List<Music>>().apply {
        value = ArrayList()
    }

    lateinit var mediaHelper: MediaHelper

    var project: Project? = null
        set(value) {
            field = value
            value?.let {
                musicSource.library = parseThemeMusic(it)

                it.story?.let {
                    storyLiveData.value = it
                    storyPeriodLiveData.value = parseStoryPeriod(it)
                    sceneLiveData.value = it.scenes[0]
                }
            }
        }

    var musicSource = MusicSource()

    private val musicPlayer = MusicPlayerV2()


    fun parseDeviceMusic(contentResolver: ContentResolver) {
        val libMusicPaths = ArrayList<String?>()
        for (libMusic in musicSource.library) {
            libMusicPaths.add(libMusic.path)
        }

        musicSource.device = DataParser.parseMusic(contentResolver, libMusicPaths)
    }

    fun importMusic(music: Music) {
        val musics: ArrayList<Music> = musicListLiveData.value as ArrayList<Music>
        project?.let {
            val musicFile = File(music.path)
            val outputFile = createMusicFile()

            outputFile?.let {
                musicFile.copyTo(it)
                music.path = it.path
            }
        }
        musics.add(music)
        musicListLiveData.value = musics
        musicLiveData.value = music
    }

    fun importEditMusic(music: Music) {
        val musics: ArrayList<Music> = musicListLiveData.value as ArrayList<Music>
        val outputFile = createMusicFile()
        outputFile?.let {
            if (music.endTime - music.startTime > storyPeriodLiveData.value ?: 0) {
                music.endTime = music.startTime + (storyPeriodLiveData.value ?: 0)
            }

            val audioEditorBuilder = AudioEditor.Builder(mediaHelper)
            audioEditorBuilder.editAudioFile = File(music.path)
            audioEditorBuilder.startTime = TimeUtils.getEncodeTimeFormt(music.startTime)
            audioEditorBuilder.endTime = TimeUtils.getEncodeTimeFormt(music.endTime)
            audioEditorBuilder.volume = music.volume

            val audioEditor = audioEditorBuilder.build()
            audioEditor.output(it)
                    .subscribeOn(Schedulers.io())
                    .observeOn(AndroidSchedulers.mainThread())
                    .doOnSubscribe {
                        showProgressingLiveData.value = true
                    }
                    .doFinally {
                        showProgressingLiveData.value = false
                    }.subscribe({
                        music.path = outputFile.path
                        musics[musics.indexOf(musicLiveData.value)] = music
                        musicListLiveData.value = musics
                        musicLiveData.value = music
                    }, {

                    })
        }
    }

    fun setMusicLoopEnable(enable: Boolean) {
        val musics: ArrayList<Music> = musicListLiveData.value as ArrayList<Music>
        musicLiveData.value?.let {
            it.isLoop = enable
            musics[musics.indexOf(it)] = it
        }
        musicListLiveData.value = musics
    }

    fun deleteMusic() {
        val musics: ArrayList<Music> = musicListLiveData.value as ArrayList<Music>
        musics.remove(musicLiveData.value)
        musicListLiveData.value = musics
        if (musics.size > 0) {
            musicLiveData.value = musics[0]
        }
    }

    fun precastMusic(music: Music) {
        musicPlayer.apply {
            this.music = music
            stop()
            init()
            start()
        }
    }

    fun playMusic(music: Music, startPosition: Int) {
        musicPlayer.apply {
            this.music = music
            playChangedListener = object : MusicPlayerV2.OnPlayChangedListener {
                override fun onPlayChanged(position: Int) {
                    musicPlayPosition.postValue(position)
                }
            }
            init()
            seekTo(startPosition)
            start()
        }
    }

    fun pauseMusic(){
        musicPlayer.pause()
    }

    fun stopMusic(){
        musicPlayer.stop()
    }

    fun setMusicVolume(volume: Float) {
        musicPlayer.setVolume(volume)
    }

    fun uploadVideo(context: Context, file: File, project: Project): Observable<UploadVideoData> {
        val id = context.getString(R.string.bomo_cloud_client_id)
        val secret = context.getString(R.string.bomo_cloud_client_secret)
        val author = project.author ?: ""
        val name = project.displayName ?: ""
        val category = project.category.id
        val filePart = MultipartBody.Part.createFormData("file", file.name, RequestBody.create(MediaType.parse("video/*"), file))
        return BoMoCloud(id, secret).videoService
                .uploadVideo(author, name, category, filePart)
    }

    fun saveProject(context: Context, project: Project, oldName: String): Completable {
        return FileUtils.saveProject(context, project, 1920, 1080)
                .andThen {
                    onProjectSavedListeners.forEach { it.run() }
                    it.onComplete()
                }
                .subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread())
    }

    private fun parseThemeMusic(project: Project): ArrayList<Music> {
        val musics = ArrayList<Music>()
        val mediaMetadataRetriever = MediaMetadataRetriever()
        val assetFolder = project.assetFolder

        project.themeAssetIndex.contains.music.values.forEach {
            val musicPath = "$assetFolder/$it"
            val file = File(musicPath)

            mediaMetadataRetriever.setDataSource(musicPath)

            val music = Music()
            music.name = file.name
            music.path = file.path
            music.duration = mediaMetadataRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_DURATION).toLong()
            music.endTime = mediaMetadataRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_DURATION).toLong()
            musics.add(music)
        }
        return musics
    }

    private fun parseStoryPeriod(story: Story): Long {
        var totalPeriod: Long = 0
        for (scene in story.scenes) {
            val period = scene.record?.period ?: 0
            totalPeriod += period
        }
        return totalPeriod
    }

    private fun createMusicFile(): File? {
        project?.let {
            val musicPath = "${FileUtils.getProjectPath(it)}/${Config.MUSIC_FOLDER_NAME}"
            val fileName = "${System.currentTimeMillis()}.mp3"
            val outputFile = File(musicPath, fileName)

            val parentFile = outputFile.parentFile
            if (!parentFile.exists() || !parentFile.isDirectory) {
                parentFile.mkdirs()
            }

            return outputFile
        }
        return null
    }

    fun onDestroy() {
        musicPlayer.release()

        project?.let {
            val musicPath = "${FileUtils.getProjectPath(it)}/${Config.MUSIC_FOLDER_NAME}"
            val musicFolder = File(musicPath)
            if (musicFolder.exists()) {
                for (file in musicFolder.listFiles()) {
                    file.delete()
                }
            }
        }
    }
}